# DISCARD IRRELEVANT COLUMNS
irrelevant_columns = {{ irrelevant_columns }}
{% if train %}
train_dataset = train_dataset.drop(irrelevant_columns, axis=1, errors="ignore")
{% endif %}
{% if test %}
test_dataset = test_dataset.drop(irrelevant_columns, axis=1, errors="ignore")
{% endif %}